// 002 u_c_server_client_socket_list
// 09 may 2005

// -- (C) Felix John COLIBRI 2004
// -- documentation: http://www.felix-colibri.com

(*$r+*)

unit u_c_server_client_socket_list;
  interface
    uses Classes, WinSock, u_c_basic_object
        , u_c_byte_buffer
        ;

    const k_tcp_ip_chunk= 1500;
    type c_server_client_socket= // one "server_client_socket"
             Class(c_basic_object)
               // -- m_name: the socket handle string
               m_socket_handle: tSocket;
               m_c_reception_buffer: c_byte_buffer;

               Constructor create_server_client_socket(p_name: String;
                   p_socket_handle: tSocket);
               function f_display_server_client_socket: String;
               function f_c_self: c_server_client_socket;

               procedure handle_fd_read_notification;
               procedure send_file(p_file_name: String);

               Destructor Destroy; Override;
             end; // c_server_client_socket

         c_server_client_socket_list= // "server_client_socket" list
             Class(c_basic_object)
               m_c_server_client_socket_list: tStringList;

               Constructor create_server_client_socket_list(p_name: String);

               function f_server_client_socket_count: Integer;
               function f_c_server_client_socket(p_server_client_socket_index: Integer): c_server_client_socket;
               function f_index_of(p_server_client_socket_handle: tSocket): Integer;
               function f_c_find_server_client_socket_by_handle(p_server_client_socket_handle: tSocket): c_server_client_socket;
               procedure add_server_client_socket(p_server_client_socket_name: String; p_c_server_client_socket: c_server_client_socket);
               function f_c_add_unique_server_client_socket(p_server_client_socket_handle: tSocket): c_server_client_socket;
               procedure delete_server_client_socket(p_server_client_socket_index: Integer);
               procedure display_server_client_socket_list;

               procedure close_server_client_socket_list;

               Destructor Destroy; Override;
             end; // c_server_client_socket_list

  implementation
    uses SysUtils, u_c_display, u_characters
        , u_winsock;

    // -- c_server_client_socket

    Constructor c_server_client_socket.create_server_client_socket(p_name: String;
        p_socket_handle: tSocket);
      begin
        Inherited create_basic_object(p_name);
        m_socket_handle:= p_socket_handle;
        m_c_reception_buffer:= c_byte_buffer.create_byte_buffer('reception', 4096);
      end; // create_server_client_socket

    function c_server_client_socket.f_display_server_client_socket: String;
      begin
        Result:= Format('%-10s ', [m_name]);
      end; // f_display_server_client_socket

    function c_server_client_socket.f_c_self: c_server_client_socket;
      begin
        Result:= Self;
      end; // f_c_self

    procedure c_server_client_socket.handle_fd_read_notification;
      var l_remaining: Integer;
          l_pt_start_reception: Pointer;
          l_packet_bytes: Integer;

          l_eol_position: Integer;
          l_file_name: String;
      begin
        display('> handle_read '+ f_socket_name(m_socket_handle));

        with m_c_reception_buffer do
        begin
          l_remaining:= m_buffer_size- m_write_index;

          // -- if not at least a tcp-ip chunk, increase the size
          if l_remaining< k_tcp_ip_chunk
            then begin
                // -- reallocate
                double_the_capacity;
                l_remaining:= m_buffer_size- m_write_index;
              end;

          // -- add the received data to the current buffer
          l_pt_start_reception:= @ m_oa_byte_buffer[m_write_index];

          // -- get the data from the client socket
          display(':Recv('+ IntToStr(m_socket_handle)+ ', )');
          l_packet_bytes:= Recv(m_socket_handle, l_pt_start_reception^, l_remaining, 0);
          if l_packet_bytes< 0
            then display_socket_error_halt(m_socket_handle,
                  'c.fd_read (Recv)', WSAGetLastError)
            else
              if l_packet_bytes> 0
                then begin
                    m_write_index:= m_write_index+ l_packet_bytes;
                    l_eol_position:= f_return_line_feed_position(0);
                    if l_eol_position> 0
                      then begin
                          // -- -3: remove eol
                          l_file_name:= f_extract_string_start_end(0, l_eol_position- 3);
                          send_file(l_file_name);
                        end;
                  end; // if l_packet_bytes
        end; // with m_c_reception_buffer
        display('< handle_read '+ f_socket_name(m_socket_handle));
      end; // handle_fd_read_notification

    procedure c_server_client_socket.send_file(p_file_name: String);
      var l_file_name: String;
          l_answer: String;
          l_result: Integer;
          l_pt_start_emission: Pointer;
      begin
        display('> send_file');
        l_file_name:= '..\socket\_data\'+ p_file_name;
        display(l_file_name);

        if FileExists(l_file_name)
          then begin
              with c_byte_buffer.create_byte_buffer('load', 0) do
              begin
                load_from_file(l_file_name);

                l_answer:= IntToStr(m_write_index)+ k_new_line;
                display(':Send(server_client_'+ IntToStr(m_socket_handle)+ ', bytes, '+
                    IntToStr(Length(l_answer))+ ', )');
                l_result:= Send(m_socket_handle, l_answer[1], Length(l_answer), 0);
                if l_result< 0
                  then display_socket_error_halt(m_socket_handle, 'c.send_string (Send)',
                       WSAGetLastError);

                l_pt_start_emission:= @ m_oa_byte_buffer[0];
                display(':Send(server_client_'+ IntToStr(m_socket_handle)+ ', bytes, '+
                    IntToStr(m_write_index)+ ', )');
                l_result:= Send(m_socket_handle, l_pt_start_emission^, m_write_index, 0);
                if l_result< 0
                  then display_socket_error_halt(m_socket_handle, 'c.send_string (Send)',
                       WSAGetLastError);
                Free;
              end; // with c_byte_buffer
            end
          else begin
              l_answer:= 'file '+ p_file_name+ ' not found'+ k_new_line;
              l_answer:= IntToStr(Length(l_answer))+ k_new_line+ l_answer;

              display(':Send(server_client_'+ IntToStr(m_socket_handle)+ ', '+ l_answer+ ', '+
                  IntToStr(Length(l_answer))+ ', )');
              l_result:= Send(m_socket_handle, l_answer[1], Length(l_answer), 0);
              if l_result< 0
                then display_socket_error_halt(m_socket_handle, 'c.send_string (Send)',
                     WSAGetLastError);
            end;
        display('< send_file');
      end; // send_file

    Destructor c_server_client_socket.Destroy;
      begin
        m_c_reception_buffer.Free;
        InHerited;
      end; // Destroy

    // -- c_server_client_socket_list

    Constructor c_server_client_socket_list.create_server_client_socket_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        m_c_server_client_socket_list:= tStringList.Create;
      end; // create_server_client_socket_line

    function c_server_client_socket_list.f_server_client_socket_count: Integer;
      begin
        Result:= m_c_server_client_socket_list.Count;
      end; // f_server_client_socket_count

    function c_server_client_socket_list.f_c_server_client_socket(p_server_client_socket_index: Integer): c_server_client_socket;
      begin
        Result:= c_server_client_socket(m_c_server_client_socket_list.Objects[p_server_client_socket_index]);
      end; //  f_c_server_client_socket

    function c_server_client_socket_list.f_index_of(p_server_client_socket_handle: tSocket): Integer;
      begin
        Result:= m_c_server_client_socket_list.IndexOf(IntToStr(p_server_client_socket_handle));
      end; // f_index_of

    function c_server_client_socket_list.f_c_find_server_client_socket_by_handle(p_server_client_socket_handle: tSocket): c_server_client_socket;
      var l_socket_handle_string: String;
          l_index_of: Integer;
      begin
        l_socket_handle_string:= IntToStr(p_server_client_socket_handle);
        l_index_of:= f_index_of(p_server_client_socket_handle);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_server_client_socket(m_c_server_client_socket_list.Objects[l_index_of]);
      end; // f_c_find_server_client_socket_by_handle

    procedure c_server_client_socket_list.add_server_client_socket(p_server_client_socket_name: String; p_c_server_client_socket: c_server_client_socket);
      begin
        m_c_server_client_socket_list.AddObject(p_server_client_socket_name, p_c_server_client_socket);
      end; // add_server_client_socket

    function c_server_client_socket_list.f_c_add_unique_server_client_socket(p_server_client_socket_handle: tSocket): c_server_client_socket;
      var l_index_of: Integer;
          l_socket_handle_string: String;
      begin
        l_index_of:= f_index_of(p_server_client_socket_handle);
        if l_index_of>= 0
          then Result:= Nil
          else begin
              l_socket_handle_string:= IntToStr(p_server_client_socket_handle);
              Result:= c_server_client_socket.create_server_client_socket(l_socket_handle_string,
                  p_server_client_socket_handle);
              add_server_client_socket(l_socket_handle_string, Result);
            end;
      end; // f_c_add_unique_server_client_socket

    procedure c_server_client_socket_list.delete_server_client_socket(p_server_client_socket_index: Integer);
      begin
        m_c_server_client_socket_list.Delete(p_server_client_socket_index);
      end; // delete_server_client_socket

    procedure c_server_client_socket_list.display_server_client_socket_list;
      var l_server_client_socket_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_server_client_socket_count));

        for l_server_client_socket_index:= 0 to f_server_client_socket_count- 1 do
          display(f_c_server_client_socket(l_server_client_socket_index).f_display_server_client_socket);
      end; // display_server_client_socket_list

    procedure c_server_client_socket_list.close_server_client_socket_list;
      var l_server_client_socket_index: Integer;
      begin
        for l_server_client_socket_index:= 0 to f_server_client_socket_count- 1 do
          with f_c_server_client_socket(l_server_client_socket_index) do
            CloseSocket(m_socket_handle);
      end; // close_server_socket_list

    Destructor c_server_client_socket_list.Destroy;
      var l_server_client_socket_index: Integer;
      begin
        for l_server_client_socket_index:= 0 to f_server_client_socket_count- 1 do
          f_c_server_client_socket(l_server_client_socket_index).Free;
        m_c_server_client_socket_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_server_client_socket_list
    end. // u_c_server_client_socket_list
